import asyncio
from collections import namedtuple
import copy
from hashlib import sha256
import time
import unittest
from uuid import uuid4, UUID

from preseem import FakeNetworkMetadataModel, NetworkMetadataReference, FakePreseemModel
from preseem.model_sync import sync_preseem_elements, sync_preseem_model, ElementSyncItem
from preseem_protobuf.model.common_pb2 import ElementStatus
from preseem_protobuf.model.account_pb2 import AccountStatus
from preseem_protobuf.model.common_pb2 import AccountType
from preseem_protobuf.model.service_pb2 import SetAttachmentReq
from preseem.source_model import Attachment, AccessPoint, Olt, Router, Service

Account = namedtuple('Account', ('id', 'type', 'status', 'name', 'first_name', 'last_name'))
Package = namedtuple('Package', ('id', 'name', 'download_rate', 'upload_rate'))

class ElementRole:
    SWITCH = UUID('41f2903a-311a-448d-9d4d-86dc8ba3cb07')
    AP = UUID('ae3f1aa5-063a-4a24-9744-822a82d107d9')
    ROUTER = UUID('bf433e28-122b-41df-b49a-77ac6544deac')
    CPE_RADIO = UUID('c72115d3-daf9-43ae-9475-083d62eae74f')
    OLT = UUID('2d9027d6-8639-40eb-8b65-b86c8ee2238b')

def has_any_fields_set(message):
    for field_desc, value in message.ListFields():
        if value:
            return True
    return False


#import logging
#logging.basicConfig(format='%(message)s', level=logging.INFO)

# XXX I came across a couple issues
# 1. same source id for different intent_roles could be possible.
# Maybe I can implicitly assume there's no overlap within a source.  Ideally.
class TestModelSync(unittest.TestCase):
    """Test the sync from nm refs to preseem model."""
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        #self.loop.set_debug(True)  # uncomment for more debug info
        asyncio.set_event_loop(self.loop)
        self.model = FakePreseemModel()
        self.nm_model = FakeNetworkMetadataModel(refs_by_type=True)
        self.company = self._await(self.model.Company.Create(name='test_1'))

    def tearDown(self):
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    async def set_ap_ref(self, id, tower, sector, ip_address=None, system=None):
        ref = NetworkMetadataReference('ap', id, {'topology2': tower, 'topology1': sector})
        if ip_address:
            ref.attributes['ip_address'] = ip_address
        if system:
            ref.attributes['system'] = system
        await self.nm_model.set_ref(ref)

    def get_ref_uuid(self, ap_id):
        ref = (self.nm_model.refs.get('ap_model') or {}).get(ap_id)
        if ref:
            uuid_str = ref.attributes.get('uuid')
            if uuid_str:
                return UUID(uuid_str)
        return None

    def test_zero_aps(self):
        """Run through with both models empty."""
        async def test():
            aps = []

            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)

            self.assertEqual(await self.model.Site.List(company_uuid=self.company.uuid), [])
            self.assertEqual(await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid), [])
            self.assertEqual(uuid_map, {})
        self._await(test())

    def test_ap_created_elem_disc_created(self):
        """A Site and ElementDiscovery are created for an AP."""
        async def test():
            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1')]

            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)

            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 1)
            site = sites[0]
            self.assertEqual(site.name, 'Site 1')
            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.name, 'AP 1')
            self.assertEqual(disc.management_ip, '192.0.2.1')
            self.assertEqual(disc.inactive, False)
            self.assertEqual(disc.site_uuid, site.uuid)
            self.assertEqual(disc.source, 'test_source')
            self.assertEqual(disc.source_id, 'ap01')
            self.assertEqual(disc.element_uuid, None)
            self.assertEqual(disc.status, None)
            self.assertEqual(uuid_map, {'ap01': UUID(bytes=disc.uuid)})
        self._await(test())

    def test_ap_removed_elem_disc_inactive(self):
        """An element is removed from the billing system."""
        async def test():
            site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            disc = await self.model.Element.CreateElementDiscovery(
                    company_uuid=self.company.uuid,
                    name='AP 1',
                    management_ip='192.0.2.1',
                    site_uuid=site.uuid,
                    source='test_source',
                    source_id='ap01',
                    intent_role_uuid=ElementRole.AP.bytes)

            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [])

            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 1)
            self.assertEqual(sites[0], site)

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.name, 'AP 1')
            self.assertEqual(disc.management_ip, '192.0.2.1')
            self.assertEqual(disc.inactive, True)
            self.assertEqual(disc.site_uuid, site.uuid)
            self.assertEqual(disc.source, 'test_source')
            self.assertEqual(disc.source_id, 'ap01')
            self.assertEqual(disc.element_uuid, None)
            self.assertEqual(disc.status, None)
            self.assertEqual(uuid_map, {})
        self._await(test())

    def test_ap_modify_attributes(self):
        """The model is updated when attributes are modified."""
        async def test():
            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1'),
                   ElementSyncItem('ap02', 'AP 2', 'Site 1', '192.0.2.2'),
                   ElementSyncItem('ap03', 'AP 3', 'Site 1', '192.0.2.3')]
            uuid_map1 = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 3)
            discs1 = {x.source_id: x for x in discs}
            aps = [ElementSyncItem('ap01', 'AP 01', 'Site 1', '192.0.2.1'),
                   ElementSyncItem('ap02', 'AP 2', 'Site 2', '192.0.2.2'),
                   ElementSyncItem('ap03', 'AP 3', 'Site 1', '192.0.2.30')]

            uuid_map2 = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)

            self.assertEqual(uuid_map1, uuid_map2)
            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 2)
            sites = {x.name: x for x in sites}
            site1 = sites.get('Site 1')
            self.assertIsNotNone(site1)
            site2 = sites.get('Site 2')
            self.assertIsNotNone(site2)
            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 3)
            discs = {x.source_id: x for x in discs}
            self.assertGreater(discs['ap01'].updated_at, discs1['ap01'].updated_at)
            self.assertEqual(discs1['ap01'].name, 'AP 1')
            self.assertEqual(discs['ap01'].name, 'AP 01')
            discs['ap01'].name = discs1['ap01'].name
            discs['ap01'].updated_at = discs1['ap01'].updated_at
            self.assertEqual(discs1['ap01'], discs['ap01'])
            self.assertGreater(discs['ap02'].updated_at, discs1['ap02'].updated_at)
            self.assertEqual(discs1['ap02'].site_uuid, site1.uuid)
            self.assertEqual(discs['ap02'].site_uuid, site2.uuid)
            discs['ap02'].site_uuid = discs1['ap02'].site_uuid
            discs['ap02'].updated_at = discs1['ap02'].updated_at
            self.assertEqual(discs1['ap02'], discs['ap02'])
            self.assertGreater(discs['ap03'].updated_at, discs1['ap03'].updated_at)
            self.assertEqual(discs1['ap03'].management_ip, '192.0.2.3')
            self.assertEqual(discs['ap03'].management_ip, '192.0.2.30')
            discs['ap03'].management_ip = discs1['ap03'].management_ip
            discs['ap03'].updated_at = discs1['ap03'].updated_at
            self.assertEqual(discs1['ap03'], discs['ap03'])
        self._await(test())

    def test_reactivation(self):
        """APs and site objects are reused."""
        async def test():
            site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            disc = await self.model.Element.CreateElementDiscovery(
                    company_uuid=self.company.uuid,
                    name='AP 1',
                    management_ip='192.0.2.1',
                    site_uuid=site.uuid,
                    source='test_source',
                    source_id='ap01',
                    intent_role_uuid=ElementRole.AP.bytes)
            await self.model.Site.Update(company_uuid=self.company.uuid, uuid=site.uuid, inactive=True)
            await self.model.Element.UpdateElementDiscovery(company_uuid=self.company.uuid, uuid=disc.uuid, inactive=True)

            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1')]
            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps, nm_model=self.nm_model)

            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 1)
            self.assertEqual(sites[0].uuid, site.uuid)
            self.assertEqual(sites[0].name, site.name)
            self.assertEqual(sites[0].inactive, False)

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            self.assertEqual(discs[0].uuid, disc.uuid)
            self.assertEqual(discs[0].name, disc.name)
            self.assertEqual(discs[0].management_ip, disc.management_ip)
            self.assertEqual(discs[0].inactive, False)
            self.assertEqual(discs[0].site_uuid, disc.site_uuid)
            self.assertEqual(discs[0].source, 'test_source')
            self.assertEqual(discs[0].source_id, 'ap01')
            self.assertEqual(discs[0].element_uuid, None)
            self.assertEqual(discs[0].status, None)
            self.assertEqual(uuid_map, {'ap01': UUID(bytes=disc.uuid)})
        self._await(test())

    def test_resolved_ap_reactivation(self):
        """An resolved AP discovery is reactivated.  This clears the element_uuid
           attribute and the elem_discovery reference.  See STM-9823."""
        async def test():
            elem_uuid = UUID('52af5281-51aa-4869-a297-7e24fe2d18d9')
            site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            disc = await self.model.Element.CreateElementDiscovery(
                    company_uuid=self.company.uuid,
                    name='AP 1',
                    management_ip='192.0.2.1',
                    site_uuid=site.uuid,
                    source='test_source',
                    source_id='ap01',
                    element_uuid=elem_uuid.bytes,
                    intent_role_uuid=ElementRole.AP.bytes)
            await self.model.Site.Update(company_uuid=self.company.uuid, uuid=site.uuid, inactive=True)
            await self.model.Element.UpdateElementDiscovery(company_uuid=self.company.uuid, uuid=disc.uuid, inactive=True)
            ref = NetworkMetadataReference('elem_discovery', str(UUID(bytes=disc.uuid)), {'element_uuid': str(elem_uuid)})
            await self.nm_model.set_ref(ref)
            self.assertEqual(len(self.nm_model.refs.get('elem_discovery', {})), 1)

            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1')]
            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps, nm_model=self.nm_model)

            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 1)
            self.assertEqual(sites[0].uuid, site.uuid)
            self.assertEqual(sites[0].name, site.name)
            self.assertEqual(sites[0].inactive, False)

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            self.assertEqual(discs[0].uuid, disc.uuid)
            self.assertEqual(discs[0].name, disc.name)
            self.assertEqual(discs[0].management_ip, disc.management_ip)
            self.assertEqual(discs[0].inactive, False)
            self.assertEqual(discs[0].site_uuid, disc.site_uuid)
            self.assertEqual(discs[0].source, 'test_source')
            self.assertEqual(discs[0].source_id, 'ap01')
            self.assertEqual(discs[0].element_uuid, None)
            self.assertEqual(discs[0].status, None)
            self.assertEqual(uuid_map, {'ap01': UUID(bytes=disc.uuid)})
            self.assertEqual(len(self.nm_model.refs.get('elem_discovery', {})), 0)
        self._await(test())


    def test_active_site_preferred(self):
        """If multiple sites with the same name exist, an active one is used."""
        async def test():
            new_site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            old_site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            await self.model.Site.Update(company_uuid=self.company.uuid, uuid=old_site.uuid, inactive=True)
            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1')]

            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.site_uuid, new_site.uuid)
        self._await(test())

    def test_empty_source_raises_error(self):
        """The source field cannot be the empty string."""
        async def test():
            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, '', ElementRole.AP, aps)
        self._await(test())

    def test_empty_source_id_raises_error(self):
        """The source_id field cannot be the empty string."""
        async def test():
            aps = [ElementSyncItem('', 'AP 1', 'Site 1', '192.0.2.1')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
        self._await(test())

    def test_empty_name_raises_error(self):
        """The name field cannot be the empty string."""
        async def test():
            aps = [ElementSyncItem('ap01', '', 'Site 1', '192.0.2.1')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
        self._await(test())

    def test_empty_site_name_raises_error(self):
        """The site_name field cannot be the empty string."""
        async def test():
            aps = [ElementSyncItem('ap01', 'AP 1', '', '192.0.2.1')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
        self._await(test())

    def test_invalid_ip_address_raises_error(self):
        """The ip_address field must be a valid IP address."""
        async def test():
            aps = [ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.256')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
        self._await(test())

    def test_duplicate_item_raises_error(self):
        """A duplicate id is not permitted."""
        async def test():
            aps = [
                    ElementSyncItem('ap01', 'AP 1', 'Site 1', '192.0.2.1'),
                    ElementSyncItem('ap01', 'AP 2', 'Site 1', '192.0.2.2')]

            with self.assertRaises(ValueError):
                uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, aps)
        self._await(test())

    def test_deactivation_only_happens_once(self):
        """Deactivation is only done once."""
        async def test():
            site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            disc = await self.model.Element.CreateElementDiscovery(
                    company_uuid=self.company.uuid,
                    name='AP 1',
                    management_ip='192.0.2.1',
                    site_uuid=site.uuid,
                    source='test_source',
                    source_id='ap01',
                    intent_role_uuid=ElementRole.AP.bytes)

            # Set it to inactive once and note the inactive timestamp
            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [])
            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.inactive, True)
            inactive_time = disc.inactive_time
            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [])

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.inactive, True)
            self.assertEqual(disc.inactive_time, inactive_time)
        self._await(test())

    def test_foreign_sources_are_ignored(self):
        """Deactivation is only done for elements matching this source."""
        async def test():
            site = await self.model.Site.Create(
                    company_uuid=self.company.uuid,
                    name='Site 1')
            disc = await self.model.Element.CreateElementDiscovery(
                    company_uuid=self.company.uuid,
                    name='AP 1',
                    management_ip='192.0.2.1',
                    site_uuid=site.uuid,
                    source='alt_source',
                    source_id='ap01',
                    intent_role_uuid=ElementRole.AP.bytes)

            uuid_map = await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [])

            sites = await self.model.Site.List(company_uuid=self.company.uuid)
            self.assertEqual(len(sites), 1)
            self.assertEqual(sites[0], site)

            discs = await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)
            self.assertEqual(len(discs), 1)
            disc = discs[0]
            self.assertEqual(disc.name, 'AP 1')
            self.assertEqual(disc.management_ip, '192.0.2.1')
            self.assertEqual(disc.inactive, False)
            self.assertEqual(disc.site_uuid, site.uuid)
            self.assertEqual(disc.source, 'alt_source')
            self.assertEqual(disc.source_id, 'ap01')
            self.assertEqual(disc.element_uuid, None)
            self.assertEqual(disc.status, None)
            self.assertEqual(uuid_map, {})
        self._await(test())

    def test_package_create(self):
        """A Package is created in the model."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkg = packages[0]
            self.assertEqual(pkg.name, 'Package 1')
            self.assertEqual(pkg.download_rate, 25000)
            self.assertEqual(pkg.upload_rate, 5000)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg01')
        self._await(test())

    def test_package_empty_source_id_raises_error(self):
        """The source_id field for a package cannot be the empty string."""
        async def test():
            pkgs = {'': Package('', 'Package 1', 25000, 5000)}
            with self.assertRaises(ValueError) as err:
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_empty_name_raises_error(self):
        """The name field cannot be the empty string."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', '', 25000, 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_non_integer_rate_raises_error(self):
        """The rate must be an integer if specified."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 'asdf')}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 'asdf', 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_negative_rate_raises_error(self):
        """The rate cannot be a negative integer."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', -1, 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, -1)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_zero_rate_is_null(self):
        """A rate of zero is sent as a null rate."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 0, 5000),
                    'pkg02': Package('pkg02', 'Package 2', None, 0)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 2)
            packages = sorted(packages, key=lambda x: x.source_id)
            pkg = packages[0]
            self.assertEqual(pkg.name, 'Package 1')
            self.assertEqual(pkg.download_rate, None)
            self.assertEqual(pkg.upload_rate, 5000)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg01')
            pkg = packages[1]
            self.assertEqual(pkg.name, 'Package 2')
            self.assertEqual(pkg.download_rate, None)
            self.assertEqual(pkg.upload_rate, None)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg02')
        self._await(test())

    def test_package_update(self):
        """A Package is updated in the model."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)

            pkgs['pkg01'] = Package('pkg01', 'Package 1a', 50000, 10000)
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkg = packages[0]
            self.assertEqual(pkg.name, 'Package 1a')
            self.assertEqual(pkg.download_rate, 50000)
            self.assertEqual(pkg.upload_rate, 10000)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg01')
        self._await(test())

    def test_package_reactivate(self):
        """An inactive package is reactivated in the model."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            await self.model.Package.Update(company_uuid=self.company.uuid, uuid=packages[0].uuid, inactive=True)
            pkgs['pkg01'] = Package('pkg01', 'Package 1a', 25000, 5000)
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkg = packages[0]
            self.assertEqual(pkg.name, 'Package 1a')
            self.assertEqual(pkg.download_rate, 25000)
            self.assertEqual(pkg.upload_rate, 5000)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg01')
        self._await(test())

    def test_package_update_empty_name_raises_error(self):
        """The name field cannot be the empty string on an update."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkgs = {'pkg01': Package('pkg01', '', 25000, 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_update_non_integer_rate_raises_error(self):
        """The rate must be an integer if specified on an update."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 'asdf')}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 'asdf', 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_update_negative_rate_raises_error(self):
        """The rate cannot be a negative integer on an update."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', -1, 5000)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, -1)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
        self._await(test())

    def test_package_update_zero_rate_is_null(self):
        """A rate of zero is sent as a null rate."""
        async def test():
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000),
                    'pkg02': Package('pkg02', 'Package 2', None, 5000)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 2)
            pkgs = {'pkg01': Package('pkg01', 'Package 1', 0, 5000),
                    'pkg02': Package('pkg02', 'Package 2', None, 0)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], pkgs)
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 2)
            packages = sorted(packages, key=lambda x: x.source_id)
            pkg = packages[0]
            self.assertEqual(pkg.name, 'Package 1')
            self.assertEqual(pkg.download_rate, None)
            self.assertEqual(pkg.upload_rate, 5000)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg01')
            pkg = packages[1]
            self.assertEqual(pkg.name, 'Package 2')
            self.assertEqual(pkg.download_rate, None)
            self.assertEqual(pkg.upload_rate, None)
            self.assertEqual(pkg.source, 'test_source')
            self.assertEqual(pkg.source_id, 'pkg02')
        self._await(test())

    def test_package_deactivation(self):
        """Deactivation works and is only done once."""
        async def test():
            pkg = await self.model.Package.Create(
                    company_uuid=self.company.uuid,
                    source='test_source',
                    source_id='pkg01',
                    name='Package 1')

            # Set it to inactive once and note the inactive timestamp
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], {})
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            package = packages[0]
            self.assertEqual(package.inactive, True)
            inactive_time = package.inactive_time
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], {})
            packages = await self.model.Package.List(company_uuid=self.company.uuid)
            self.assertEqual(len(packages), 1)
            package = packages[0]
            self.assertEqual(package.inactive, True)
            self.assertEqual(package.inactive_time, inactive_time)
        self._await(test())

    def test_account_create(self):
        """An Account is created in the model."""
        async def test():
            accounts = {'acct01': Account('acct01', AccountType.ACCOUNT_TYPE_RESIDENTIAL, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Name01', None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.type, AccountType.ACCOUNT_TYPE_RESIDENTIAL)
            self.assertEqual(acct.status, AccountStatus.ACCOUNT_STATUS_ACTIVE)
            self.assertEqual(acct.name, 'Name01')
            self.assertEqual(acct.first_name, None)
            self.assertEqual(acct.last_name, None)
        self._await(test())

    def test_account_empty_source_id_raises_error(self):
        """The source_id field for an account cannot be the empty string."""
        async def test():
            accounts = {'': Account('', AccountType.ACCOUNT_TYPE_RESIDENTIAL, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Name01', None, None)}
            with self.assertRaises(ValueError) as err:
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_type_is_optional(self):
        """The Type field is not required."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Name01', None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.type, None)

    def test_account_type_wrong_type_raises_error(self):
        """The Type field must be an integer type."""
        async def test():
            accounts = {'acct01': Account('acct01', 'ACCOUNT_TYPE_RESIDENTIAL', AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Name01', None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_type_bad_value_raises_error(self):
        """The Type field must be a postivite integer."""
        async def test():
            accounts = {'acct01': Account('acct01', AccountStatus.ACCOUNT_STATUS_INVALID, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Name01', None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_status_is_required(self):
        """The Status field is required."""
        async def test():
            accounts = {'acct01': Account('acct01', None, None, 'Name01', None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_status_wrong_type_raises_error(self):
        """The Status field must be an integer type."""
        async def test():
            accounts = {'acct01': Account('acct01', None, 'ACCOUNT_STATUS_ACTIVE', 'Name01', None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_status_bad_value_raises_error(self):
        """The Status field must be a postivite integer."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_INVALID, 'Name01', None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_name_is_optional(self):
        """The name field is not required."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.name, None)

    def test_account_name_wrong_type_raises_error(self):
        """The name field must be a string type."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 12334, None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_name_empty_string_raises_error(self):
        """The name field must not be an empty string."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, "", None, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_first_name_is_optional(self):
        """The first_name field is not required."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.first_name, None)

    def test_account_first_name_wrong_type_raises_error(self):
        """The first_name field must be a string type."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, 12334, None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_first_name_empty_string_raises_error(self):
        """The first_name field must not be an empty string."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, "", None)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_last_name_is_optional(self):
        """The last_name field is not required."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.last_name, None)

    def test_account_last_name_wrong_type_raises_error(self):
        """The last_name field must be a string type."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, None, 12334)}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_last_name_empty_string_raises_error(self):
        """The last_name field must not be an empty string."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, None, "")}
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
        self._await(test())

    def test_account_update(self):
        """An Account is updated in the model."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)

            accounts['acct01'] = Account('acct01', AccountType.ACCOUNT_TYPE_RESIDENTIAL, AccountStatus.ACCOUNT_STATUS_ACTIVE, None, 'Test', 'Account')
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.type, AccountType.ACCOUNT_TYPE_RESIDENTIAL)
            self.assertEqual(acct.status, AccountStatus.ACCOUNT_STATUS_ACTIVE)
            self.assertEqual(acct.name, None)
            self.assertEqual(acct.first_name, 'Test')
            self.assertEqual(acct.last_name, 'Account')
            self.assertTrue(acct.updated_at > acct.created_at)
        self._await(test())

    def test_account_deactivation(self):
        """Deactivation works and is only done once."""
        async def test():
            pkg = await self.model.Account.Create(
                    company_uuid=self.company.uuid,
                    source='test_source',
                    source_id='acct01',
                    status=AccountStatus.ACCOUNT_STATUS_ACTIVE)

            # Set it to inactive once and note the inactive timestamp
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts={})
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            account = accounts[0]
            self.assertEqual(account.inactive, True)
            inactive_time = account.inactive_time
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts={})
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            account = accounts[0]
            self.assertEqual(account.inactive, True)
            self.assertEqual(account.inactive_time, inactive_time)
        self._await(test())

    def test_account_reactivate(self):
        """An inactive account is reactivated in the model."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            await self.model.Account.Update(company_uuid=self.company.uuid, uuid=accounts[0].uuid, inactive=True)
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            accounts = await self.model.Account.List(company_uuid=self.company.uuid)
            self.assertEqual(len(accounts), 1)
            acct = accounts[0]
            self.assertEqual(acct.inactive, False)
            self.assertEqual(acct.type, None)
            self.assertEqual(acct.status, AccountStatus.ACCOUNT_STATUS_ACTIVE)
            self.assertEqual(acct.name, 'Account 1')
            self.assertEqual(acct.source, 'test_source')
            self.assertEqual(acct.source_id, 'acct01')
        self._await(test())

    async def checkSourceService(self, ssvc):
        accounts = {}
        for account in await self.model.Account.List(company_uuid=self.company.uuid):
            if account.source_id in accounts:
                if isinstance(accounts[account.source_id], list):
                    accounts[account.source_id].append(account)
                else:
                    accounts[account.source_id] = [accounts[account.source_id], account]
            else:
                accounts[account.source_id] = account
        packages = {}
        for package in await self.model.Package.List(company_uuid=self.company.uuid):
            if package.source_id in packages:
                if isinstance(packages[package.source_id], list):
                    packages[package.source_id].append(package)
                else:
                    packages[package.source_id] = [packages[package.source_id], package]
            else:
                packages[package.source_id] = package
        services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
        msvc = services.get(ssvc.id)
        self.assertIsNotNone(msvc)
        self.assertEqual(msvc.source_id, ssvc.id)
        self.assertEqual(msvc.download_rate, ssvc.download_rate)
        self.assertEqual(msvc.upload_rate, ssvc.upload_rate)
        self.assertEqual(msvc.subscriber_identifier, ssvc.subscriber_identifier)
        self.assertEqual(msvc.source_url, ssvc.url)
        if ssvc.account_id:
            account = accounts.get(ssvc.account_id)
            if isinstance(account, list):  # multiple matching Account objects
                account = next((x for x in account if x.source==msvc.source), None)
                if account is None:  # no matching sources
                    self.assertTrue(ssvc.errors)
            if account or not ssvc.errors:
                self.assertIsNotNone(account)
                self.assertEqual(account.uuid, msvc.account_uuid)
        if ssvc.package_id:
            package = packages.get(ssvc.package_id)
            if isinstance(package, list):  # multiple matching Package objects
                package = next((x for x in package if x.source==msvc.source), None)
                if package is None:  # no matching sources
                    self.assertTrue(ssvc.errors)
            if package or not ssvc.errors:
                self.assertIsNotNone(package)
                self.assertEqual(package.uuid, msvc.package_uuid)
        if ssvc.attachment:
            if has_any_fields_set(ssvc.attachment.pb):
                self.assertEqual(ssvc.attachment.pb, msvc.attachment)
            else:
                self.assertIsNone(msvc.attachment)

    def test_service_create(self):
        """A Service is created in the model."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            service.attachment.imsi = '310150123456789'
            service.attachment.cpe_serial_number = 'S12345'
            service.attachment.cpe_name = '123:456'
            service.attachment.device_mac_address = '00:0A:95:9D:68:FF'
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            service.attachment.network_prefixes.append('192.0.2.101')
            service.attachment.network_prefixes.append('3ffe:1900:9876::/48')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_no_account(self):
        """If a service references an account that does not exist, account_uuid is not set."""
        async def test():
            service = Service('service01', account_id='acct01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            with self.assertRaises(AssertionError):
                await self.checkSourceService(service)
            service.account_id = None
            await self.checkSourceService(service)

    def test_service_no_package(self):
        """If a service references a package that does not exist, package_uuid is not set."""
        async def test():
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts)
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            with self.assertRaises(AssertionError):
                await self.checkSourceService(service)
            service.package_id = None
            await self.checkSourceService(service)
        self._await(test())

    def test_service_account_collision_resolved(self):
        """If multiple accounts with matching source_id exist, the one with the same source as the service is matched."""
        async def test():
            acct1 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='test_source',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            acct2 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='source2',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            acct3 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='source3',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            service = Service('service01', account_id='acct01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_account_collision_error(self):
        """If multiple accounts with matching source_id exist and none match source, an error is flagged on the service."""
        async def test():
            acct1 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='source1',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            acct2 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='source2',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            acct3 = await self.model.Account.Create(
                company_uuid=self.company.uuid,
                source='source3',
                source_id='acct01',
                status=AccountStatus.ACCOUNT_STATUS_ACTIVE)
            service = Service('service01', account_id='acct01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_package_collision_resolved(self):
        """If multiple packages with matching source_id exist, the one with the same source as the service is matched."""
        async def test():
            pkg1 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='test_source',
                source_id='pkg01')
            pkg2 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='source2',
                source_id='pkg01')
            pkg3 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='source3',
                source_id='pkg01')
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_package_collision_error(self):
        """If multiple packages with matching source_id exist and none match source, an error is flagged on the service."""
        async def test():
            pkg1 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='source1',
                source_id='pkg01')
            pkg2 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='source2',
                source_id='pkg01')
            pkg3 = await self.model.Package.Create(
                company_uuid=self.company.uuid,
                source='source3',
                source_id='pkg01')
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_update(self):
        """A Service is updated in the model."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            service.attachment.imsi = '310150123456789'
            service.attachment.cpe_serial_number = 'S12345'
            service.attachment.cpe_name = '123:456'
            service.attachment.device_mac_address = '00:0A:95:9D:68:FF'
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            service.attachment.network_prefixes.append('192.0.2.101')
            service.attachment.network_prefixes.append('3ffe:1900:9876::/48')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)

            # change something at each level
            service.upload_rate = 21000
            service.subscriber_identifier = None
            service.url = 'https://my.new.source.url'
            service.attachment.cpe_name = None
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            await self.checkSourceService(service)
        self._await(test())

    def test_service_update_error(self):
        """Updating a field resolved by the syncer to an invalid value leaves the previous value."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id=None, download_rate=100000, upload_rate=20000, subscriber_identifier='account01')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services)
            await self.checkSourceService(service)

            service.package_id = 'non-existent-package'
            service.download_rate = 101000
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            with self.assertRaises(AssertionError):
                await self.checkSourceService(service)
            service.package_id = None
            await self.checkSourceService(service)
        self._await(test())

    def test_service_deactivation(self):
        """A service is deactivated if it is not passed to the syncer."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id=None, download_rate=100000, upload_rate=20000, subscriber_identifier='account01')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services)
            await self.checkSourceService(service)

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services={})
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            await self.checkSourceService(service)
            msvc = services.get(service.id)
            self.assertIsNotNone(msvc)
            self.assertEqual(msvc.inactive, True)
            t = msvc.inactive_time

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services={})
            await self.checkSourceService(service)
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            msvc = services.get(service.id)
            self.assertIsNotNone(msvc)
            self.assertEqual(msvc.inactive_time, t)
        self._await(test())

    def test_service_deactivation_disabled(self):
        """Service deactivation can be disabled to leave old services created."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id=None, download_rate=100000, upload_rate=20000, subscriber_identifier='account01')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services, deactivate=False)
            await self.checkSourceService(service)

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services={}, deactivate=False)
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            await self.checkSourceService(service)
            msvc = services.get(service.id)
            self.assertIsNotNone(msvc)
            self.assertEqual(msvc.inactive, False)
            self.assertEqual(msvc.inactive_time, None)
        self._await(test())

    def test_service_reactivation(self):
        """A service can be reactivated."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            service = Service('service01', account_id='acct01', package_id=None, download_rate=100000, upload_rate=20000, subscriber_identifier='account01')
            service.attachment.cpe_mac_address = '00:01:02:03:04:05'
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, services=services)
            await self.checkSourceService(service)

            # Deactivate it
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services={})
            await self.checkSourceService(service)
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            msvc = services.get(service.id)
            self.assertIsNotNone(msvc)
            self.assertEqual(msvc.inactive, True)

            # Reactivate it
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            await self.checkSourceService(service)
            services = {x.source_id: x for x in await self.model.Service.List(company_uuid=self.company.uuid)}
            msvc = services.get(service.id)
            self.assertIsNotNone(msvc)
            self.assertEqual(msvc.inactive, False)
        self._await(test())

    def test_service_ip_prefix_format(self):
        """If the service returns a different IP format, the comparison is done correctly."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.network_prefixes.append('192.0.2.101')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)

            # change the prefix that comes back to a "/32" (reflects actual behavior)
            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            msvc = model_services[0]
            self.model._services[msvc.company_uuid][msvc.uuid].attachment.network_prefixes[0] += '/32'

            # The sync shouldn't change anything in the model.
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            self.assertEqual(self.model._services[msvc.company_uuid][msvc.uuid].attachment.network_prefixes[0], "192.0.2.101/32")
        self._await(test())

    def test_service_null_attachment(self):
        """If the service attachment was not set, it can be set by the sync."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            msvc = model_services[0]

            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
        self._await(test())

    def test_service_empty_attachment(self):
        """Comparing an empty attachment is handled properly."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            msvc = model_services[0]

            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            # check that we did not actually do an update
            self.assertEqual(msvc.updated_at, model_services[0].updated_at)
        self._await(test())

    def test_service_mac_order(self):
        """The order of MAC addresses doesn't matter."""
        async def test():
            # create a service
            service = Service('service01')
            services = {'service01': service}
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            service.attachment.mac_addresses.append('06:07:08:09:0a:0b')

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            await self.checkSourceService(service)

            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            msvc = model_services[0]

            service = Service('service01')
            services = {'service01': service}
            service.attachment.mac_addresses.append('06:07:08:09:0a:0b')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], services=services)
            with self.assertRaises(AssertionError):
                await self.checkSourceService(service)  # will fail due to order
            model_services = await self.model.Service.List(company_uuid=self.company.uuid)
            # check that we did not actually do an update
            self.assertEqual(msvc.updated_at, model_services[0].updated_at)
        self._await(test())

    def test_service_update_attachment_config(self):
        """A Service can have its cpe_mac field ignored based on configuration."""
        async def test():
            # create a service
            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
            service.attachment.mac_addresses.append('00:01:02:03:04:05')
            service.attachment.network_prefixes.append('192.0.2.101')
            services = {'service01': service}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)
            service.attachment.pb.CopyFrom(service.attachment.pb)

            # change the CPE MAC address externally.
            svc = (await self.model.Service.List(company_uuid=self.company.uuid))[0]
            set_att_req = SetAttachmentReq()
            set_att_req.attachment.CopyFrom(svc.attachment)
            set_att_req.attachment.cpe_mac_address = "00:01:00:01:00:01"
            svc = await self.model.Service.Update(company_uuid=self.company.uuid, uuid=svc.uuid, attachment_req=set_att_req)

            # by default, it gets changed back.
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
            await self.checkSourceService(service)  # fails since set externally

            # change the CPE MAC address externally.
            svc = (await self.model.Service.List(company_uuid=self.company.uuid))[0]
            set_att_req = SetAttachmentReq()
            set_att_req.attachment.CopyFrom(svc.attachment)
            set_att_req.attachment.cpe_mac_address = "00:01:00:01:00:01"
            svc = await self.model.Service.Update(company_uuid=self.company.uuid, uuid=svc.uuid, attachment_req=set_att_req)

            # sync, ignoring this field so a different system is able to change it.
            # (something else has to change, too, or it won't modify the attachment.)
            service.attachment.username = "user02"
            cfg = {'service': {'attachment': {'cpe_mac_address': {'ignore': True}}}}

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services, cfg=cfg)
            with self.assertRaises(AssertionError):
                await self.checkSourceService(service)  # fails since set externally
        self._await(test())

    async def checkSourceElement(self, sobj, role_uuid, deleted=False):
        sites = {}
        for site in await self.model.Site.List(company_uuid=self.company.uuid):
            if site.name in sites:
                if isinstance(sites[site.name], list):
                    sites[site.name].append(site)
                else:
                    sites[site.name] = [sites[site.name], site]
            else:
                sites[site.name] = site
        mobjs = {x.source_id: x for x in await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid) if x.intent_role_uuid == role_uuid.bytes}
        mobj = mobjs.get(sobj.id)
        self.assertIsNotNone(mobj)
        self.assertEqual(mobj.source_id, sobj.id)
        self.assertEqual(mobj.name, sobj.name)
        self.assertEqual(mobj.management_ip, sobj.host)
        self.assertEqual(mobj.inactive, deleted)
        if sobj.site:
            site = sites.get(sobj.site)
            if site or not sobj.errors:
                self.assertIsNotNone(site)
                self.assertEqual(site.uuid, mobj.site_uuid)
        self.assertEqual(mobj.uuid, sobj._uuid)

    async def checkSourceRouter(self, sobj, deleted=False):
        return await self.checkSourceElement(sobj, ElementRole.ROUTER, deleted=deleted)

    async def checkSourceOlt(self, sobj, deleted=False):
        return await self.checkSourceElement(sobj, ElementRole.OLT, deleted=deleted)

    async def checkSourceAP(self, sobj, deleted=False):
        return await self.checkSourceElement(sobj, ElementRole.AP, deleted=deleted)

    def test_router_create(self):
        """An Router is created in the model."""
        async def test():
            router = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1', url='https://myisp.com/routers/router01')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_no_site(self):
        """If a router does not specify a site, it is still created."""
        async def test():
            router = Router('router01', host='192.0.2.1', name='Router 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_update(self):
        """A Router is updated in the model."""
        async def test():
            # create a router
            router = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)

            # change attributes
            router.name = 'Router 01'
            router.host = '192.0.2.254'
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_update(self):
        """A Router is updated in the model."""
        async def test():
            # create a router
            router = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)

            # change attributes
            router.name = 'Router 01'
            router.host = '192.0.2.254'
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_update_error(self):
        """Updating a field resolved by the syncer to an invalid value leaves the previous value."""
        async def test():
            # create a router
            router = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)

            # test internal field checking
            router.name = ''
            self.assertEqual(router.name, 'Router 1')
            self.assertTrue(router.errors)
            router.errors.pop()
            router.host = ''
            self.assertEqual(router.host, '192.0.2.1')
            self.assertTrue(router.errors)
            router.errors.pop()

            # test field checking done by sync
            router.name = None
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            router.host = None
            with self.assertRaises(ValueError):
                await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)

            router.name = 'Router 1'
            router.host = '192.0.2.1'
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_uuid(self):
        """The router source object is populated with its uuid after sync."""
        async def test():
            router = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1', url='https://myisp.com/routers/router01')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceRouter(router)
        self._await(test())

    def test_router_uuid_source(self):
        """PPA-1176 Do not set the UUID to a discovery object of another source."""
        async def test():
            router1 = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1', url='https://myisp.com/routers/router01')
            router2 = Router('router01', host='192.0.2.1', name='Router 1', site='Site 1', url='https://myisp.com/routers/router01')
            routers = {'router01': router1}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            routers = {'router01': router2}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source_2', ElementRole.AP, [], routers=routers)
            # sync the first one again to trigger the bug
            routers = {'router01': router1}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)

            mobjs = {x.source: x for x in await self.model.Element.ListElementDiscovery(company_uuid=self.company.uuid)}
            self.assertEqual(len(mobjs), 2)
            mobj1 = mobjs.pop('test_source')
            mobj2 = mobjs.pop('test_source_2')
            self.assertIsNotNone(mobj1)
            self.assertIsNotNone(mobj2)
            self.assertEqual(mobj1.uuid, router1._uuid)
            self.assertEqual(mobj2.uuid, router2._uuid)
        self._await(test())

    def test_olt(self):
        """OLTs are handled properly in the model, the same way routers are."""
        async def test():
            # create an OLT
            olt = Olt('olt01', host='192.0.2.1', name='OLT 1', site='Site 1')
            olts = {'olt01': olt}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], olts=olts)
            await self.checkSourceOlt(olt)

            # add a router so we can test any interactions between them
            router = Router('router01', host='192.0.2.254', name='Router 1', site='Site 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], routers=routers)
            await self.checkSourceOlt(olt)
            await self.checkSourceRouter(router)

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], olts=olts, routers=routers)
            await self.checkSourceOlt(olt)
            await self.checkSourceRouter(router)

            # Delete the router and update the OLT
            olt.host = '192.0.2.2'
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], olts=olts, routers={})
            await self.checkSourceOlt(olt)
            await self.checkSourceRouter(router, deleted=True)
        self._await(test())

    def test_ap(self):
        """APs are handled properly the "modern" way in the model, the same way routers are."""
        async def test():
            # create an AP
            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1')
            aps = {'ap01': ap}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', None, [], aps=aps)
            await self.checkSourceAP(ap)

            # add a router so we can test any interactions between them
            router = Router('router01', host='192.0.2.254', name='Router 1', site='Site 1')
            routers = {'router01': router}
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', None, [], routers=routers)
            await self.checkSourceAP(ap)
            await self.checkSourceRouter(router)

            await sync_preseem_model(self.model, self.company.uuid, 'test_source', None, [], aps=aps, routers=routers)
            await self.checkSourceAP(ap)
            await self.checkSourceRouter(router)

            # Delete the router and update the AP
            ap.host = '192.0.2.2'
            await sync_preseem_model(self.model, self.company.uuid, 'test_source', None, [], aps=aps, routers={})
            await self.checkSourceAP(ap)
            await self.checkSourceRouter(router, deleted=True)
        self._await(test())

    async def checkSourceElement2(self, sobj):
        """Check actual elements synced from source objects."""
        sites = {}
        for site in await self.model.Site.List(company_uuid=self.company.uuid):
            if site.name in sites:
                if isinstance(sites[site.name], list):
                    sites[site.name].append(site)
                else:
                    sites[site.name] = [sites[site.name], site]
            else:
                sites[site.name] = site
        mobjs = {x.source_id: x for x in await self.model.Element.List(company_uuid=self.company.uuid)}
        mobj = mobjs.get(sobj.id)
        self.assertIsNotNone(mobj)
        self.assertEqual(mobj.source_id, sobj.id)
        self.assertEqual(mobj.name, sobj.name)
        self.assertEqual(mobj.management_ip, sobj.host)
        self.assertEqual(mobj.inactive, False)
        self.assertEqual(mobj.status, ElementStatus.ELEMENT_STATUS_UNKNOWN)
        self.assertEqual(mobj.poller_hash, sobj.poller_hash)
        if sobj.site:
            site = sites.get(sobj.site)
            if site or not sobj.errors:
                self.assertIsNotNone(site)
                self.assertEqual(site.uuid, mobj.site_uuid)
        self.assertEqual(mobj.uuid, sobj._uuid)

    def test_element_sync(self):
        """Elements can be directly created and updated."""
        async def test():
            # Create two elements.
            ap_poller_hash = sha256(b"ap1").digest()
            olt_poller_hash = sha256(b"olt1").digest()
            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
            elems = {'ap01': ap, 'olt01': olt}
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            # Update the elements.
            ap.host = "192.0.2.2"
            olt.site = None
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
            self.assertEqual(len(mobjs), 2)
        self._await(test())

    def test_element_sync_hardware_swap(self):
        """Hardware swaps are handled by updating the correct poller hash element."""
        async def test():
            ap_poller_hash_1 = sha256(b"ap1").digest()
            ap_poller_hash_2 = sha256(b"ap2").digest()
            olt_poller_hash = sha256(b"olt1").digest()
            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash_1)
            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
            elems = {'ap01': ap, 'olt01': olt}
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            # The AP gets swapped for new hardware.
            ap.poller_hash = ap_poller_hash_2
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
            self.assertEqual(len(mobjs), 3)
        self._await(test())

    def test_element_sync_source_id_change(self):
        """If the source ID changes for the same hardware, it should be the same element."""
        async def test():
            ap_poller_hash = sha256(b"ap1").digest()
            olt_poller_hash = sha256(b"olt1").digest()
            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
            elems = {'ap01': ap, 'olt01': olt}
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            # The AP ID get changed
            ap = AccessPoint('ap02', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
            elems = {'ap02': ap, 'olt01': olt}
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)
            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
            self.assertEqual(len(mobjs), 2)

            # Change something else
            ap.name = "AP 2"
            elems = {'ap02': ap, 'olt01': olt}
            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
            await self.checkSourceElement2(olt)
            await self.checkSourceElement2(ap)

            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
            self.assertEqual(len(mobjs), 2)
        self._await(test())



import asyncio
from collections import namedtuple
from enum import Enum
import logging
import socket
from sys import intern
from time import monotonic

import pytricia

from preseem_grpc_model import network_metadata_pb2
from preseem.grpc_client import PreseemGrpcClient

NetworkMetadataEntity = namedtuple('NetworkMetadataEntity',
                                   ('network_prefix', 'start', 'attributes'))
NetworkMetadataReference = namedtuple('NetworkMetadataReference',
                                      ('type', 'value', 'attributes'))
NetworkMetadataIndex = namedtuple('NetworkMetadataIndex', ('type', 'attribute'))
NetworkMetadataLowercaseIndex = namedtuple('NetworkMetadataLowercaseIndex',
                                           ('type', 'attribute'))
Op = Enum('Operation', 'SET DEL SYN')


class NetworkMetadataClient(PreseemGrpcClient):

    def __init__(self, api_key, host=None, port=None, beta=None):
        super().__init__(api_key,
                         host or 'nmapi{}.preseem.com'.format('-beta' if beta else ''),
                         port,
                         compression=True)


def ip_sort_key(x):
    """Sort key for listing IP address strings."""
    try:
        return socket.inet_aton(x)
    except OSError:
        return bytes(0)


def _resolve_refs(refs, attrs):
    "Helper function to resolve reference attributes." ""
    # first resolve non-refs, then recurse alphabetically on refs.  first one wins.
    # - first resolve on current ref
    # - then resolve refs, alphabetically, and overwrite only if not set in map.
    ref_attrs = []
    for t, v in list(attrs.items()):
        if isinstance(v, Reference):
            ref_attrs.append((t, v))

    for t, v in sorted(ref_attrs):
        refmap = refs.get(t)
        if refmap:
            ref = refmap.get(v)
            if ref:
                child_attrs = _resolve_refs(refs, ref.attributes.copy())
                for at, av in child_attrs.items():
                    if at not in attrs:
                        attrs[at] = av
    return attrs


def _diff(src, dst, refs_by_type=False):
    """Helper function to return operations to make dst like src"""
    adds = []
    dels = []
    mods = []
    if refs_by_type:
        keys = sorted(set([(x[0], y) for x in src.items() for y in x[1].keys()] + \
                          [(x[0], y) for x in dst.items() for y in x[1].keys()]))
    else:
        keys = sorted(set(list(src.keys()) + list(dst.keys())))
    for key in keys:
        if refs_by_type:
            dobj = (dst.get(key[0]) or {}).get(key[1])
            sobj = (src.get(key[0]) or {}).get(key[1])
        else:
            dobj = dst.get(key)
            sobj = src.get(key)
        if sobj is None:
            dels.append(dobj)
        elif dobj is None:
            adds.append(sobj)
        elif dobj != sobj:
            mods.append(sobj)
    return adds, dels, mods


class Reference(str):
    """This is used to hint to the GRPC code that it should create a ref"""
    pass


class uint64(int):
    """This is used to hint to the GRPC code that it should create a uint"""
    pass


def _attrs_to_dict(attrs):
    result = {}
    for attr, data in attrs.items():
        if data.type == network_metadata_pb2.NetworkMetadataAttribute.REFERENCE:
            val = Reference(data.refVal.value)
        elif data.type == network_metadata_pb2.NetworkMetadataAttribute.BOOL:
            val = data.boolVal
        elif data.type == network_metadata_pb2.NetworkMetadataAttribute.DOUBLE:
            val = data.doubleVal
        elif data.type == network_metadata_pb2.NetworkMetadataAttribute.STRING:
            val = data.stringVal
        elif data.type == network_metadata_pb2.NetworkMetadataAttribute.INT64:
            val = data.int64Val
        elif data.type == network_metadata_pb2.NetworkMetadataAttribute.UINT64:
            val = data.uint64Val
        result[intern(attr)] = val
    return result


def _nm_set_attrs(nm_attrs, py_attrs):
    for t, v in py_attrs.items():
        if v is None:
            continue  # ignore null values
        apb = nm_attrs[t]
        apb.name = t
        ty = type(v)
        if ty is Reference:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.REFERENCE
            apb.refVal.type = t
            apb.refVal.value = str(v)
        elif ty is bool:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.BOOL
            apb.boolVal = v
        elif ty is float:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.DOUBLE
            apb.doubleVal = v
        elif ty is str:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.STRING
            apb.stringVal = v
        elif ty is int:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.INT64
            apb.int64Val = v
        elif ty is uint64:
            apb.type = network_metadata_pb2.NetworkMetadataAttribute.UINT64
            apb.uint64Val = v
        else:
            raise RuntimeError('unknown attr type {}'.format(ty.__name__))


class PreseemNetworkMetadataModel(object):

    def __init__(
            self,
            client,
            company_id=None,
            cbk=None,
            refs_by_type=None,  # deprecated
            indexes=None):
        """
        Pass in refs_by_type=True to use the newer API where the "refs"
        attribute is {type: {id: obj}}.  Default is {(type, id): obj}.
        Indexes are specified here as a list of NetworkMetadataIndex.
        """
        self._client = client
        self.company_id = company_id  # used for multi-company keys
        self._cbk = cbk  # optional callback for updates
        self.refs_by_type = refs_by_type
        self.refs = {}  # public: current references
        self.nms = {}  # public: current entities
        self.refs_last_updated = {}  # ref_type -> eventloop.time() of last change
        self.pt4 = pytricia.PyTricia(32)
        self.syn = None  # dict of GRPC objects we receive in synchronize state
        self.nm_syn = None
        self.start_load = None
        self._ent_sub_task = None
        self._ref_sub_task = None
        self._ents_loaded = asyncio.Condition()
        self._refs_loaded = asyncio.Condition()
        self._ents_lock = asyncio.Lock()
        self._refs_lock = asyncio.Lock()
        self._index = {x.type: {} for x in indexes or []}
        self._index_cfg = {x.type: {} for x in indexes or []}
        for index in indexes or []:
            self._index[index.type][index.attribute] = {}
            self._index_cfg[index.type][index.attribute] = index

    def close(self):
        self._client.close_subscriptions()
        if self._ent_sub_task:
            self._ent_sub_task.cancel()
            self._ent_sub_task = None
        if self._ref_sub_task:
            self._ref_sub_task.cancel()
            self._ref_sub_task = None

    async def handle_ent_update(self, msg, syn=False):
        if syn:
            self.nm_syn = {}
        if msg.action == network_metadata_pb2.NetworkMetadataEntityAction.LOADED:
            self.nms = self.nm_syn
            self.pt4 = pytricia.PyTricia(32)
            for ip, nm in self.nms.items():
                if ':' not in ip:
                    try:
                        self.pt4[ip] = nm
                    except Exception as err:
                        logging.info('Error setting nm entity %s: %s', ip, err)
                        continue
            self.nm_syn = None
            if self._cbk:
                for nm in self.nms.values():
                    try:
                        await self._cbk(nm, deleted=False)
                    except Exception:
                        pass
            async with self._ents_loaded:
                self._ents_loaded.notify()
        elif msg.action == network_metadata_pb2.NetworkMetadataEntityAction.SET:
            nm = self._make_nm(msg.entity)
            if self.nm_syn is None:
                self.nms[nm.network_prefix] = nm
                if ':' not in nm.network_prefix:
                    try:
                        self.pt4[nm.network_prefix] = nm
                    except Exception as err:
                        logging.info('Error setting nm entity %s: %s',
                                     nm.network_prefix, err)
                if self._cbk:
                    try:
                        await self._cbk(nm, deleted=False)
                    except Exception:
                        pass
            else:
                self.nm_syn[nm.network_prefix] = nm
                if ':' not in msg.entity.network_prefix:
                    try:
                        self.pt4[nm.network_prefix] = nm
                    except Exception as err:
                        logging.info('Error setting nm entity %s: %s',
                                     nm.network_prefix, err)
                        del self.nm_syn[nm.network_prefix]
        elif msg.action == network_metadata_pb2.NetworkMetadataEntityAction.DELETE:
            nm = self._make_nm(msg.entity)
            try:
                del self.nms[nm.network_prefix]
            except KeyError:
                pass
            try:
                del self.pt4[nm.network_prefix]
            except KeyError:
                pass
            if self._cbk:
                try:
                    await self._cbk(nm, deleted=True)
                except Exception:
                    pass

    def is_loading(self):
        """Returns true if we are currently loading references."""
        return self.start_load is not None

    def ref_handler(self, iterator):
        """New generator approach.  This is called for the duration of a connection, in a separate worker thread."""
        syn = {}
        start_load = monotonic()
        while True:
            msg = iterator.next()
            if msg.action == network_metadata_pb2.NetworkMetadataReferenceEntityAction.SET:
                ref = NetworkMetadataReference(intern(msg.entity.type),
                                               msg.entity.value,
                                               _attrs_to_dict(msg.entity.attributes))
                if syn is None:
                    yield Op.SET, ref
                else:
                    refs = syn.get(ref.type)
                    if refs is None:
                        refs = syn[ref.type] = {}
                    curt = self.refs.get(ref.type)
                    if curt:
                        cur = curt.get(ref.value)
                        if cur == ref:
                            ref = cur
                    refs[ref.value] = ref
            elif msg.action == network_metadata_pb2.NetworkMetadataReferenceEntityAction.DELETE:
                ref = NetworkMetadataReference(intern(msg.entity.type),
                                               msg.entity.value,
                                               _attrs_to_dict(msg.entity.attributes))
                if syn is None:
                    yield Op.DEL, ref
                else:
                    # this should not happen; if it does should I delete from syn?
                    print("Got delete during startup!")

            elif msg.action == network_metadata_pb2.NetworkMetadataReferenceEntityAction.LOADED:
                load_time = round(monotonic() - start_load)
                if load_time >= 5:  # log a long load time
                    logging.info("Loaded nm refs in %s seconds", load_time)
                yield Op.SYN, syn
                syn = None
            else:
                print("INVALID ACTION", msg.action)

    async def handle_ref_update2(self, op, ref):
        now = asyncio.get_event_loop().time()
        if op == Op.SET:
            refs = self.refs.get(ref.type)
            if refs is None:
                refs = self.refs[ref.type] = {}
            cur = refs.get(ref.value)
            refs[ref.value] = ref
            index = self._index.get(ref.type)
            if index is not None:
                for ia, ir in index.items():
                    if ia == '_LOWER':
                        refs = ir.get(None)
                        if refs is None:
                            refs = ir[None] = {}
                        refs[ref.value.lower()] = ref
                        continue
                    is_lower = isinstance(self._index_cfg[ref.type][ia],
                                          NetworkMetadataLowercaseIndex)
                    if cur:
                        # Remove the old ref from any indexes
                        attr = cur.attributes.get(ia)
                        if attr:
                            if is_lower:
                                attr = attr.lower()
                            refs = ir.get(attr)
                            if refs:
                                refs.pop(ref.value, None)
                    attr = ref.attributes.get(ia)
                    if attr:
                        if is_lower:
                            attr = attr.lower()
                        refs = ir.get(attr)
                        if refs is None:
                            refs = ir[attr] = {}
                        refs[ref.value] = ref
            if self._cbk:
                try:
                    await self._cbk(ref, deleted=False)
                except Exception:
                    pass
            self.refs_last_updated[ref.type] = now

        elif op == Op.DEL:
            try:
                # use the saved ref so we have its attributes
                ref = self.refs[ref.type].pop(ref.value)
                if not self.refs[ref.type]:
                    del self.refs[ref.type]
                index = self._index.get(ref.type)
                if index is not None:
                    for ia, ir in index.items():
                        is_lower = isinstance(self._index_cfg[ref.type][ia],
                                              NetworkMetadataLowercaseIndex)
                        if ia == '_LOWER':
                            refs = ir.get(None)
                            if refs:
                                refs.pop(ref.value.lower(), None)
                            continue
                        attr = ref.attributes.get(ia)
                        if attr:
                            if is_lower:
                                attr = attr.lower()
                            refs = ir.get(attr)
                            if refs:
                                refs.pop(ref.value, None)
                self.refs_last_updated[ref.type] = now
            except KeyError:
                pass
            if self._cbk:
                try:
                    await self._cbk(ref, deleted=True)
                except Exception:
                    pass

        elif op == Op.SYN:
            self.refs = ref
            self.refs_last_updated = {ty: now for ty in self.refs}
            for it, index in self._index.items():
                for ia, ir in index.items():
                    ir.clear()
                    if ia == '_LOWER':  # lowercase value index
                        ir[None] = {
                            n.lower(): r
                            for n, r in (self.refs.get(it) or {}).items()
                        }
                        continue
                    is_lower = isinstance(self._index_cfg[it][ia],
                                          NetworkMetadataLowercaseIndex)
                    for ref in (self.refs.get(it) or {}).values():
                        attr = ref.attributes.get(ia)
                        if attr:
                            if is_lower:
                                attr = attr.lower()
                            refs = ir.get(attr)
                            if refs is None:
                                refs = ir[attr] = {}
                            refs[ref.value] = ref
            if self._cbk:
                if self.refs_by_type:
                    refs = [
                        v for d in list(self.refs.values()) for v in list(d.values())
                    ]
                else:
                    refs = self.refs.values()
                for ref in refs:
                    try:
                        await self._cbk(ref, deleted=False)
                    except Exception:
                        pass
            async with self._refs_loaded:
                self._refs_loaded.notify()
        else:
            print("INVALID ACTION")

    def _make_nm(self, msg, resolve_refs=False):
        cidr = msg.network_prefix.split('/')
        net = cidr[0] if len(cidr) == 1 or cidr[1] == '32' else msg.network_prefix
        obj = NetworkMetadataEntity(net, msg.start.seconds,
                                    _attrs_to_dict(msg.attributes))
        if resolve_refs:
            _resolve_refs(refs, obj.attributes)
        return obj

    def get_ref_index(self, reftype, attr, val=None):
        """Get a map of reference types of a type with a given attr value."""
        return self._index[reftype][attr].get(val)

    async def list_nm(self, resolve_refs=False):
        result = []

        refs = {}
        if resolve_refs:
            for ref in await self.list_refs():
                refmap = refs.get(ref.type)
                if refmap is None:
                    refmap = refs[ref.type] = {}
                refmap[ref.value] = ref

        for msg in await self._client.nm_list(company_id=self.company_id):
            result.append(self._make_nm(msg, resolve_refs))
        return result

    async def set_nm(self, entity):
        assert isinstance(entity, NetworkMetadataEntity)
        ts = network_metadata_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()
        ts.FromNanoseconds(int(entity.start * 1000000000))
        obj = network_metadata_pb2.NetworkMetadataEntity(
            address_space='default', network_prefix=entity.network_prefix, start=ts)
        _nm_set_attrs(obj.attributes, entity.attributes)
        await self._client.nm_set(obj, company_id=self.company_id)

    async def del_nm(self, entity):
        assert isinstance(entity, NetworkMetadataEntity)
        ts = network_metadata_pb2.google_dot_protobuf_dot_timestamp__pb2.Timestamp()
        ts.FromNanoseconds(int(entity.start * 1000000000))
        obj = network_metadata_pb2.NetworkMetadataEntity(
            address_space='default', network_prefix=entity.network_prefix, start=ts)
        _nm_set_attrs(obj.attributes, entity.attributes)
        await self._client.nm_del(obj, company_id=self.company_id)

    async def list_refs(self):
        result = []
        for msg in await self._client.nm_ref_list(company_id=self.company_id):
            ref = NetworkMetadataReference(intern(msg.type), msg.value,
                                           _attrs_to_dict(msg.attributes))
            result.append(ref)
        return result

    async def set_ref(self, ref):
        assert isinstance(ref, NetworkMetadataReference)
        obj = network_metadata_pb2.NetworkMetadataReferenceEntity(type=ref.type,
                                                                  value=ref.value)
        _nm_set_attrs(obj.attributes, ref.attributes)
        await self._client.nm_ref_set(obj, company_id=self.company_id)

    async def del_ref(self, ref):
        assert isinstance(ref, NetworkMetadataReference)
        obj = network_metadata_pb2.NetworkMetadataReferenceEntity(type=ref.type,
                                                                  value=ref.value)
        await self._client.nm_ref_del(obj, company_id=self.company_id)

    async def _load_ents(self):
        async with self._ents_lock:
            if self._ent_sub_task is None:
                cor = self._client.nm_ent_subscribe(self.handle_ent_update,
                                                    company_id=self.company_id)
                task = asyncio.get_event_loop().create_task(cor)
                async with self._ents_loaded:
                    await self._ents_loaded.wait()
                self._ent_sub_task = task

    async def _load_refs(self):
        async with self._refs_lock:
            if self._ref_sub_task is None:
                cor = self._client.nm_ref_subscribe(self.handle_ref_update2,
                                                    company_id=self.company_id,
                                                    handler=self.ref_handler)
                task = asyncio.get_event_loop().create_task(cor)
                async with self._refs_loaded:
                    await self._refs_loaded.wait()
                self._ref_sub_task = task

    async def update(self):
        """Start the model; sync with the central database.  Return when we are
           synchronized with the starting state."""
        tasks = []
        if self._ref_sub_task is None:
            tasks.append(self._load_refs())
        if self._ent_sub_task is None:
            tasks.append(self._load_ents())
        if tasks:
            await asyncio.gather(*tasks)

    async def copy(self, dst_model):
        """Perform operations to make a destination model the same as this."""

        async def copy_refs():
            adds, dels, mods = _diff(self.refs, dst_model.refs, self.refs_by_type)
            logging.info('refs: {} adds, {} dels, {} mods'.format(
                len(adds), len(dels), len(mods)))
            for obj in adds:
                await dst_model.set_ref(obj)
            for obj in dels:
                await dst_model.del_ref(obj)
            for obj in mods:
                await dst_model.set_ref(obj)

        async def copy_nms():
            adds, dels, mods = _diff(self.nms, dst_model.nms)
            logging.info('ipas: {} adds, {} dels, {} mods'.format(
                len(adds), len(dels), len(mods)))
            for obj in adds:
                await dst_model.set_nm(obj)
            for obj in dels:
                await dst_model.del_nm(obj)
            for obj in mods:
                await dst_model.set_nm(obj)

        await asyncio.gather(self.update(), dst_model.update())
        await asyncio.gather(copy_refs(), copy_nms())

    def diff(self, rhs_model):
        """Compare this model to another one."""

        def resolve_refs(refs, attrs):
            """Resolve references.  return the combined attribute map."""
            result = {}
            bfs = [attrs]
            while bfs:
                for t, v in sorted(bfs.pop(0).items()):
                    if v is None:
                        continue  # just ignore attrs set to None
                    if isinstance(v, Reference):
                        if self.refs_by_type:
                            ref = (refs.get(t) or {}).get(v)
                        else:
                            ref = refs.get((t, v))
                        if ref:
                            bfs.append(ref.attributes)
                    else:
                        result[t] = result.get(t, v)
            return result

        diffs = []
        for key in sorted(set(list(self.nms.keys()) + list(rhs_model.nms.keys())),
                          key=ip_sort_key):
            lobj = self.nms.get(key)
            robj = rhs_model.nms.get(key)
            if lobj:
                lattrs = resolve_refs(self.refs, lobj.attributes)
            if robj:
                rattrs = resolve_refs(rhs_model.refs, robj.attributes)
            if robj is None and lattrs:
                diffs.append((key, lattrs, None))
            elif lobj is None and rattrs:
                diffs.append((key, None, rattrs))
            elif robj and lobj and lattrs != rattrs:
                diffs.append((key, {
                    x: y
                    for x, y in lattrs.items() if rattrs.get(x) != y
                }, {
                    x: y
                    for x, y in rattrs.items() if lattrs.get(x) != y
                }))
        return diffs



"""UISP Integration"""
import json
from . import HttpClient


class UnmsApiClient(HttpClient):
    """Query the uNMS API."""

    def __init__(self, url, api_token, verbose=0):
        super().__init__(url + '/v2.1', verify_ssl=False)
        self._hdr = {'X-Auth-Token': api_token}
        self.verbose = verbose

    async def query(self, path):
        """Make a GET request to an API endpoint."""
        r = await self.request(path, headers=self._hdr)
        if self.verbose > 2:
            print(f'uNMS Request: {path}')
            if self.verbose > 2:
                print(json.dumps(r, indent=2))
        return r

    async def authenticate(self):
        """Do a simple query to check authentication."""
        await self.get_nms_version()

    async def get_sites(self):
        return await self.query('/sites')

    async def get_devices(self, site_id=None):
        params = '?siteId={}'.format(site_id) if site_id else ''
        return await self.query('/devices{}'.format(params))

    async def get_device_interfaces(self, device_id):
        return await self.query(f'/devices/{device_id}/interfaces')

    async def get_nms_settings(self):
        return await self.query('/nms/settings')

    async def get_nms_version(self):
        data = await self.query('/nms/version')
        if data and isinstance(data, dict):
            return data.get('version')
        raise RuntimeError(f'Failed to get NMS version: {data}')

    async def get_credentials(self, device_id):
        """Get credentials for a device from the UISP Vault.  Requires admin token."""
        return await self.query(f'/vault/{device_id}/credentials')



"""
Tests for http2 functionality in http_client.
We are really just unit testing the replay function here, but real test
scripts were used to generate these, which are:
  test/data/http2.01.py
  test/data/http2.02.py
"""
import asyncio
import json
import unittest

from preseem.http_client import ReplayHttpClient

class TestHttp2(unittest.TestCase):
    """Tests for HTTP2 client replay."""
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        asyncio.set_event_loop(self.loop)

    async def _make_client(self, url, file):
        return ReplayHttpClient(url, file)

    def make_client(self, url, file):
        return self._await(self._make_client(url, file))

    def tearDown(self):
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    async def _get_stream_content(self, r):
        content = b''
        async for chunk in r.aiter_bytes():
            content += chunk
        return content

    def test_request(self):
        """Test replay of a http2 request"""
        client = self.make_client('https://ipinfo.io', 'test/data/http2.01.http2')
        r = self._await(client.http2_request('/', headers={'accept': 'application/json'}))
        self.assertEqual(r.headers, {
            'access-control-allow-origin': '*',
            'alt-svc': 'h3=":443"; ma=2592000,h3-29=":443"; ma=2592000',
            'content-encoding': 'gzip',
            'content-type': 'application/json; charset=utf-8',
            'date': 'Tue, 20 Sep 2022 01:35:34 GMT',
            'referrer-policy': 'strict-origin-when-cross-origin',
            'strict-transport-security': 'max-age=2592000; includeSubDomains',
            'vary': 'Accept-Encoding',
            'via': '1.1 google',
            'x-content-type-options': 'nosniff',
            'x-envoy-upstream-service-time': '3',
            'x-frame-options': 'SAMEORIGIN',
            'x-xss-protection': '1; mode=block',
        })
        self.assertEqual(json.loads(r.content.decode('utf-8')), 
        {
            "ip": "216.58.123.77",
            "hostname": "216-58-123-77.cpe.distributel.net",
            "city": "Kitchener",
            "region": "Ontario",
            "country": "CA",
            "loc": "43.4260,-80.5438",
            "org": "AS11814 Distributel Communications Limited",
            "postal": "N2N",
            "timezone": "America/Toronto",
            "readme": "https://ipinfo.io/missingauth",
        })
        self._await(client.close())

    def test_stream(self):
        """Test replay of a http2 stream"""
        client = self.make_client('https://ipinfo.io', 'test/data/http2.02.http2')
        r = self._await(client.http2_stream('/', headers={'accept': 'application/json'}))
        content = self._await(self._get_stream_content(r))
        self.assertEqual(r.headers, {
            'access-control-allow-origin': '*',
            'alt-svc': 'h3=":443"; ma=2592000,h3-29=":443"; ma=2592000',
            'content-encoding': 'gzip',
            'content-type': 'application/json; charset=utf-8',
            'date': 'Tue, 20 Sep 2022 01:42:41 GMT',
            'referrer-policy': 'strict-origin-when-cross-origin',
            'strict-transport-security': 'max-age=2592000; includeSubDomains',
            'vary': 'Accept-Encoding',
            'via': '1.1 google',
            'x-content-type-options': 'nosniff',
            'x-envoy-upstream-service-time': '4',
            'x-frame-options': 'SAMEORIGIN',
            'x-xss-protection': '1; mode=block',
        })
        self.assertEqual(json.loads(content.decode('utf-8')), 
        {
            "ip": "216.58.123.77",
            "hostname": "216-58-123-77.cpe.distributel.net",
            "city": "Kitchener",
            "region": "Ontario",
            "country": "CA",
            "loc": "43.4260,-80.5438",
            "org": "AS11814 Distributel Communications Limited",
            "postal": "N2N",
            "timezone": "America/Toronto",
            "readme": "https://ipinfo.io/missingauth",
        })
        self._await(client.close())



"""
Common classes used by this package.
"""
import encodings.idna  # workaround for PPA-918 (python 3.7 bug)
import asyncio
from collections import namedtuple
from icmplib import async_ping
import ipaddress
import logging
import os
import time

import aiohttp
import aiohttp.web
from packaging import version

from .http_client import HttpClient
from .mikrotik import MikrotikApiError
from .ssh_client import SshClient
import asyncssh  # has to come after ssh_client import to suppress warnings
from prometheus_client.exposition import generate_latest

# Set an IPv6 enabled interface for utility code to use.
# Functions (i.e. parse_ip_value) will check its "ipv6_enabled" var.
_ipv6_settings = None


def set_ipv6_settings(obj):
    global _ipv6_settings
    _ipv6_settings = obj


_ping_sem = None  # used to prevent too many open files error


async def ping(host):
    """Ping a host.  Returns rtt if we can ping it, 0.0 otherwise."""
    global _ping_sem
    if not _ping_sem:
        _ping_sem = asyncio.Semaphore(200)
    try:
        async with _ping_sem:
            result = await async_ping(host, count=1, privileged=False)
        return result.avg_rtt
    except asyncio.CancelledError:
        raise
    except Exception as err:
        logging.warning('%s error pinging host %s: %s', type(err).__name__, host, err)


STANDARD_PORT_METHOD = {
    80: 'http',
    443: 'https',
}


async def http_probe(host,
                     http_client_cls,
                     methods=('https', 'http'),
                     ports=None,
                     **kwargs):
    """
    Probe a host to see if it's listening for http(s).
    Returns a http_client_cls object if it is.
    This requires the caller to pass in the HttpClient class (for unit testing)
    """

    async def handler(response):
        """
        Handler to process the HTTP response.
        Returns True if a valid response is detected.
        """
        if 200 <= response.status <= 400:  # accept a redirect, its listening.
            if response.status == 302 and response.request_info.url.scheme == 'http':
                loc = response.headers.get('location')
                if loc and loc.startswith(
                        'https'):  # redirect to https, don't return http
                    return False
            return True
        return False

    # Build URLs to probe.
    urls = [
        f"{method}://{host}{f':{port}' if port and STANDARD_PORT_METHOD.get(port) != method else ''}"
        for method in methods for port in ports or [None]
    ]

    clients = [
        http_client_cls(url,
                        verify_ssl=False,
                        conn_timeout=5,
                        read_timeout=5,
                        connection_limit=len(urls),
                        **kwargs) for url in urls
    ]
    result = None
    try:
        responses = await asyncio.gather(*[
            client.request('', num_retries=0, allow_redirects=False, handler=handler)
            for client in clients
        ],
                                         return_exceptions=True)
        for client, response in zip(clients, responses):
            if response is True and not result:  # valid response and result not yet set
                result = client
    except asyncio.CancelledError:
        result = None  # we are cancelled; let the finally clause clean up
        raise
    finally:
        await asyncio.gather(
            *[client.close() for client in clients if client is not result])
    return result


def normalize_mac(mac):
    """Try to return a normalized MAC address from input."""
    mac_address = None
    mac = mac.strip()
    if mac:
        try:
            s = '%012x' % int(mac, 16)
            mac_address = ':'.join(a + b for a, b in zip(s[::2], s[1::2]))
        except (TypeError, ValueError):  # t probably has some delimiters
            try:
                m = mac.replace('-', ':').replace('.',
                                                  ':').replace(' ',
                                                               ':').replace("'", '')
                mac_address = ':'.join(['%02x' % int(x, 16) for x in m.split(':')])
            except (TypeError, ValueError):
                raise ValueError('Invalid MAC address "{}"'.format(mac))
        if mac_address is None or len(mac_address) != 17:
            raise ValueError('Invalid MAC address "{}"'.format(mac))
    return mac_address


def parse_ip_value(val, max_width=None, strict=False):
    """Parse an IP address or CIDR and return a list of IP address strings."""
    ipv6_enabled = _ipv6_settings.ipv6_enabled if _ipv6_settings else None
    if max_width is None:
        max_width = 20 if ipv6_enabled else 24  # STM-5079
    ip_addrs = []
    if val:
        for ip in val.split(','):
            if ip.strip() == '':
                continue
            if '-' in ip:  # range
                ip_range = ip.split('-')
                if len(ip_range) != 2:
                    raise ValueError('Bad IP range: {}'.format(ip))
                try:
                    ip1 = ipaddress.ip_address(ip_range[0])
                    try:
                        ip2 = ipaddress.ip_address(ip_range[1])
                    except ValueError:  # handle range for just last octet
                        ipd = int(ip_range[1][1:]) if ip_range[1][0] == '.' else int(
                            ip_range[1])
                        ipl = ip_range[0].split('.')
                        ipl[3] = str(ipd)
                        ip2 = ipaddress.ip_address('.'.join(ipl))
                except ValueError:
                    raise ValueError('Bad IP range addresses: {}'.format(ip))
                try:
                    for net in ipaddress.summarize_address_range(ip1, ip2):
                        if ipv6_enabled:
                            ip_addrs.append(str(net))
                        else:
                            ip_addrs.extend([str(x) for x in net])
                except ValueError as e:
                    raise ValueError('Bad IP range {}: {}'.format(ip, e))
                continue
            ip_subnet = ip.lstrip().split(' ')[0].strip().split('/')
            if len(ip_subnet) < 2 or ip_subnet[1] in ('32', '255.255.255.255'):
                try:
                    ip = ipaddress.ip_address(ip_subnet[0])
                    if ip.compressed == '0.0.0.0':
                        raise ValueError('Bad IP address: {}'.format(ip_subnet[0]))
                except ValueError:
                    raise ValueError('Invalid IP address: {}'.format(val))
                ip_addrs.append(ip_subnet[0])
            else:
                try:
                    net = ipaddress.ip_network('/'.join(ip_subnet), strict=strict)
                except ValueError as e:
                    raise ValueError('Error parsing IP network {} ({})'.format(ip, e))
                if net.prefixlen >= max_width:
                    if ipv6_enabled or net.version == 6:  # STM-4192 map IPv6 subnets
                        ip_addrs.append(str(net))
                    else:
                        ip_addrs.extend([str(x) for x in net.hosts()])
                        if str(net.network_address) not in ip_addrs:
                            ip_addrs.append(str(net.network_address))  # STM-3988
                else:
                    raise ValueError('Error parsing subnet {}, too big'.format(ip))
    return ip_addrs


class TelemetryApiClient():
    PRESEEM_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

    def __init__(self, api_key, http_client_cls=HttpClient, company_id=None):
        self._client = http_client_cls('https://api.preseem.com/v1/telemetry',
                                       username=api_key)
        self.company_id = company_id

    async def close(self):
        await self._client.close()

    async def query(self, query, company=None):
        """Execute a raw query."""
        query = {'q': query}
        company = company or self.company_id
        if company:
            query['company'] = company
        try:
            data = await self._client.request('/query', params=query)
        except aiohttp.ClientResponseError as err:
            if err.code == 404:  # no stats present
                return []
            raise
        return data or []

    async def get_events(self, measurement, query, t0=None, t1=None):
        if t0:
            query['t0'] = time.strftime(self.PRESEEM_DATETIME_FORMAT, t0)
        if t1:
            query['t1'] = time.strftime(self.PRESEEM_DATETIME_FORMAT, t1)
        if self.company_id:
            query['company'] = self.company_id
        try:
            data = await self._client.request('/events/{}'.format(measurement),
                                              params=query)
        except aiohttp.ClientResponseError as err:
            if err.code == 404:  # no stats present
                return []
            raise
        if data:
            # use first and last because they could have different cols (asc+desc)
            name_map = {x: x.replace('-', '_') for x in set(data[0]) | set(data[-1])}
            nt = namedtuple(measurement, sorted(name_map.values()))
            return [nt(**{p: x.get(f) for f, p in name_map.items()}) for x in data]
        return []

    async def get_series(self, measurement, query, t0=None, t1=None):
        if t0:
            query['t0'] = time.strftime(self.PRESEEM_DATETIME_FORMAT, t0)
        if t1:
            query['t1'] = time.strftime(self.PRESEEM_DATETIME_FORMAT, t1)
        if self.company_id:
            query['company'] = self.company_id
        try:
            data = await self._client.request('/series/{}'.format(measurement),
                                              params=query)
        except aiohttp.ClientResponseError as err:
            if err.code == 404:  # no stats present
                return []
            raise
        for result in data.get('results') or []:
            # only support a single result and assume all series same cols
            r = {}
            series = result.get('series')
            if not series:
                return None
            Row = namedtuple(series[0]['name'], series[0]['columns'])
            tags = series[0].get('tags')
            if tags:
                Dataset = namedtuple('Dataset_{}'.format(series[0]['name']),
                                     sorted(tags))
            for s in series:
                rows = [Row(*row) for row in s['values']]
                if tags:
                    dataset = Dataset(*[x[1] for x in sorted(s['tags'].items())])
                    r[dataset] = rows
                else:
                    return rows
            return r
        return None


class PreseemApiKey(object):
    """Wrapper class for API key information from preseem-web."""

    def __init__(self, api_key, token):
        self.api_key = api_key
        self._hdr = {'Authorization': f'token {token}'}
        self._loaded = None
        self._data = {}
        self._api = None

    async def close(self):
        await self._api.close()
        self._api = None

    @property
    def company_id(self):
        return self._data.get('company_id')

    @property
    def company_name(self):
        return self._data.get('company_name')

    @property
    def company_uuid(self):
        return self._data.get('company_uuid')

    async def load(self):
        if self._api is None:
            self._api = HttpClient('https://app.preseem.com', verify_ssl=False)

        now = asyncio.get_event_loop().time()
        if self._loaded and self._loaded + 3600 > now:
            return
        self._data = await self._api.request('/api/v2/api-key-lookup/',
                                             headers=self._hdr,
                                             params={'api_key': self.api_key})
        self._loaded = now


class uint64(int):
    """This is used to hint to the GRPC code that it should create a uint"""


class counter32(int):
    """Helper class to manage 32-bit counter math"""

    def __init__(self):
        self.reset()

    def __int__(self):
        return self.value

    def reset(self):
        """Reset this counter to 0."""
        self.value = 0
        self.prev = None

    def update(self, val):
        """update the current value with a new value."""
        if self.prev is None:
            pass  # leave value at 0 for the initial set
        elif val < self.prev:  # handle rollover
            self.value += (0xffffffff - self.prev) + int(val)
        else:  # typical case, counter incremented
            self.value += int(val) - self.prev
        self.prev = int(val)


class Counter32Sum():
    """Helper class to manage both 32-bit counter math and counter resets"""

    def __init__(self):
        self.reset()

    def __int__(self):
        return self.value

    def reset(self):
        """Reset this counter to 0."""
        self.value = 0  #summed value
        self.prev = 0  #previous raw ctr value used

    def update(self, value, isreset=False):
        """update the current value with a new value."""
        val = int(value)
        if val < self.prev:
            #handle counter rollover or reset
            if isreset:
                # counter reset
                self.value += val
            else:
                #counter rollover
                #add +1 to account for 0xffffffff -> 0
                self.value += (0xffffffff - self.prev) + val + 1
        else:
            # typical case, counter incremented
            self.value += val - self.prev
        self.prev = val


async def get_node_versions(telemetry_api):
    """Get a dict of {node_id -> version} from the telemetry API."""
    data = {}
    rows = await telemetry_api.get_events('platform', {
        'limit': '1000,desc',
        'fields': 'node,netdev-manager'
    })
    for row in rows:
        data[row.node] = row
    return {x[0]: version.parse(x[1].netdev_manager.split()[0]) for x in data.items()}


def running_in_container() -> bool:
    """Check different hints to see if we're running in a container."""
    return os.getpid() == 1 or os.path.exists('/.dockerenv')


class MetricsServer(object):
    """Run a basic server for metrics and keepalives."""

    def __init__(self, port=9150):
        self.port = port
        self.app_metrics = aiohttp.web.Application()
        self.app_metrics.router.add_route('GET', '/metrics', self.get)

    async def start(self):
        self.server = await asyncio.get_event_loop().create_server(
            self.app_metrics.make_handler(access_log=None), '0.0.0.0', self.port)

    async def get(self, request):
        return aiohttp.web.Response(body=generate_latest(), content_type="text/plain")


def run_tool_script(coroutine):
    """Run a Presem tool script - setup the environment properly etc."""

    async def run(coroutine):
        """Async wrapper to run the tool script."""
        try:
            await coroutine
        except asyncio.CancelledError:
            pass
        except Exception as err:
            print(f'Error: {err}')

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.INFO)
    loop = asyncio.get_event_loop()
    task = loop.create_task(coroutine)
    try:
        loop.run_until_complete(task)
        task.result()
    except SystemExit:
        pass
    except KeyboardInterrupt:
        for t in asyncio.Task.all_tasks():
            t.cancel()
    except Exception as err:
        print("ERR", err)
    finally:
        try:
            task.result()
        except BaseException:
            pass
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


async def _mikrotik_probe(host,
                          mikrotik_api_cls,
                          creds,
                          raise_errors,
                          num_retries=1,
                          **kwargs):
    """Probe the mikrotik API of a host using (username, password) tuples."""
    client = None
    last_err = None
    if creds:
        mikrotik_api = mikrotik_api_cls(host, None, None, **kwargs)
        try:
            for retry_num in range(num_retries + 1):
                if retry_num > 0:
                    await asyncio.sleep(2**retry_num)
                try:
                    await asyncio.wait_for(mikrotik_api.connect(), 5.0)
                    for cred in creds:
                        last_err = None
                        mikrotik_api._user = cred[0]
                        mikrotik_api._pass = cred[1]
                        try:
                            await asyncio.wait_for(mikrotik_api.login(), 15.0)
                        except MikrotikApiError as err:
                            last_err = err
                            continue  # no match to username/password
                        except asyncio.CancelledError:
                            raise
                        except Exception as err:
                            last_err = err
                            await mikrotik_api.close()
                            break
                        client = mikrotik_api
                        return client
                except (MikrotikApiError, OSError, asyncio.TimeoutError) as err:
                    last_err = err
        except asyncio.CancelledError:
            # PPA-1020 If this coroutine gets cancelled, it can mean one of two things.
            # 1. We've been cancelled by mikrotik_api_probe because it already got
            #    a successful probe result.
            # 2. The overall probe for the element has been cancelled.  In that case,
            #    it's important for the caller (mikrotik_api_probe) to get the
            #    cancellation, so we do this eventloop yield to make sure we don't
            #    process the cancellation and return before it can handle it.
            await asyncio.sleep(0)
            last_err = None
        finally:
            if not client:
                await mikrotik_api.close()
            if last_err and raise_errors:
                raise last_err
    return client


async def mikrotik_api_probe(host,
                             mikrotik_api_cls,
                             creds,
                             raise_errors=False,
                             **kwargs):
    """Try to connect to a Mikrotik host using a list of (username, password) credentials."""
    pend = [
        asyncio.create_task(x) for x in [
            _mikrotik_probe(
                host, mikrotik_api_cls, creds, raise_errors, enable_ssl=x, **kwargs)
            for x in (True, False)
        ]
    ]
    client = None
    while pend:
        done, pend = await asyncio.wait(pend, return_when=asyncio.FIRST_COMPLETED)
        while done:
            task = done.pop()
            try:
                result = task.result()
            except asyncio.CancelledError:
                # The entire probe operation is cancelled.  Clean up all probes.
                [x.cancel() for x in done]
                [x.cancel() for x in pend]
                raise
            except:
                if raise_errors and not client and not done and not pend:
                    raise  # this is the last result, all results were errors.
                result = None
            if result:
                if client:
                    await result.close()
                else:
                    client = result
        if client:
            while pend:
                task = pend.pop()
                # PPA-1020 cancel the probe task and do not do an "await" to wait for
                # it to clean up and complete.  This is important: if we await on the
                # task and _this_ coroutine gets cancelled, the CancelledError only
                # gets caught by the probe task and it suppresses it, and we lose the
                # cancellation which is bad.  Instead, we've made sure the probe does
                # its cleanup independently and does not need to be waited on here.
                # This goes against general asyncio best practices, but seems to be a
                # reliable way to make this work in all cases.
                task.cancel()
    return client


async def _ssh_probe(host, port, ssh_client_cls, creds, raise_errors, **kwargs):
    """Probe the SSL server of a host using (username, password) tuples."""
    client = None
    last_err = None
    for cred in creds:
        last_err = None
        ssh_client = ssh_client_cls(host,
                                    port=port,
                                    username=cred[0],
                                    password=cred[1],
                                    **kwargs)
        try:
            await asyncio.wait_for(ssh_client._open(), 5.0)
            client = ssh_client
            break
        except asyncssh.PermissionDenied as err:
            last_err = err
            await ssh_client.close()
        except (asyncssh.Error, OSError, asyncio.TimeoutError) as err:
            last_err = err
            await ssh_client.close()
            break  # connectivity error, no point in trying other creds
        except Exception:
            await ssh_client.close()
            raise
    if last_err and raise_errors:
        raise last_err
    return client


async def ssh_probe(host,
                    ssh_client_cls,
                    creds,
                    port=None,
                    raise_errors=False,
                    **kwargs):
    """Try to connect to a SSH server using list of (username, password) credentials."""
    # it's done this way to support multiple ports in the future
    pend = [
        asyncio.create_task(x) for x in
        [_ssh_probe(host, port, ssh_client_cls, creds, raise_errors, **kwargs)]
    ]
    client = None
    while pend:
        done, pend = await asyncio.wait(pend, return_when=asyncio.FIRST_COMPLETED)
        while done:
            task = done.pop()
            result = task.result()
            if result:
                if client:
                    await result.close()
                else:
                    client = result
        if client:
            while pend:
                task = pend.pop()
                task.cancel()
                try:
                    await task  # let the task cleanup
                except asyncio.CancelledError:
                    pass
    return client



"""
Multi-process tests for RingBuffer to verify IPC functionality.

These tests verify the actual use case: multiple processes communicating
via shared memory ring buffer.
"""
import unittest
import multiprocessing
import time
import random
import string

from preseem.ringbuffer import RingBuffer, SHARED_MEMORY_AVAILABLE


def writer_process(buffer_name, messages, delay=0):
    """Writer process that writes messages to the ring buffer."""
    ring = RingBuffer(name=buffer_name, capacity=1024*1024, create=True)
    try:
        for msg in messages:
            if delay:
                time.sleep(delay)
            ring._write(msg)
    finally:
        ring.close()


def reader_process(buffer_name, expected_count, result_queue, timeout=5):
    """Reader process that reads messages from the ring buffer."""
    ring = RingBuffer(name=buffer_name)
    messages = []
    start = time.time()

    try:
        while len(messages) < expected_count:
            if time.time() - start > timeout:
                break
            msg = ring._read()
            if msg:
                messages.append(msg)
            else:
                time.sleep(0.01)  # Small delay if buffer empty
    finally:
        ring.close()
        result_queue.put(messages)


@unittest.skipIf(not SHARED_MEMORY_AVAILABLE, "RingBuffer requires Python 3.8+ (shared_memory not available)")
class TestRingBufferMultiProcess(unittest.TestCase):

    def setUp(self):
        """Generate unique buffer name for test isolation."""
        self.buffer_name = "test_mp_" + "".join(random.choices(string.ascii_lowercase, k=8))

    def tearDown(self):
        """Clean up shared memory."""
        try:
            ring = RingBuffer(name=self.buffer_name)
            ring.close()
            ring.unlink()
        except:
            pass  # May already be cleaned up

    def test_single_writer_single_reader(self):
        """Test basic IPC: one writer, one reader."""
        messages = [b"msg1", b"msg2", b"msg3", b"msg4", b"msg5"]
        result_queue = multiprocessing.Queue()

        # Start writer
        writer = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, messages)
        )
        writer.start()

        # Give writer time to create buffer and write
        time.sleep(0.1)

        # Start reader
        reader = multiprocessing.Process(
            target=reader_process,
            args=(self.buffer_name, len(messages), result_queue)
        )
        reader.start()

        # Wait for completion
        writer.join(timeout=5)
        reader.join(timeout=5)

        # Verify results
        self.assertFalse(writer.is_alive(), "Writer process hung")
        self.assertFalse(reader.is_alive(), "Reader process hung")

        received = result_queue.get(timeout=1)
        self.assertEqual(len(received), len(messages), "Not all messages received")
        self.assertEqual(received, messages, "Messages don't match")

    def test_writer_survives_reader_restart(self):
        """Test that writer can continue if reader stops/restarts."""
        messages_batch1 = [b"batch1_msg1", b"batch1_msg2"]
        messages_batch2 = [b"batch2_msg1", b"batch2_msg2"]
        result_queue = multiprocessing.Queue()

        # Start writer with first batch
        writer1 = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, messages_batch1)
        )
        writer1.start()
        writer1.join(timeout=5)

        # Read first batch
        reader1 = multiprocessing.Process(
            target=reader_process,
            args=(self.buffer_name, len(messages_batch1), result_queue)
        )
        reader1.start()
        reader1.join(timeout=5)

        batch1_received = result_queue.get(timeout=1)
        self.assertEqual(batch1_received, messages_batch1)

        # Write second batch (reader stopped)
        writer2 = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, messages_batch2)
        )
        writer2.start()
        writer2.join(timeout=5)

        # Read second batch with new reader
        reader2 = multiprocessing.Process(
            target=reader_process,
            args=(self.buffer_name, len(messages_batch2), result_queue)
        )
        reader2.start()
        reader2.join(timeout=5)

        batch2_received = result_queue.get(timeout=1)
        self.assertEqual(batch2_received, messages_batch2)

    def test_buffer_persistence_across_processes(self):
        """Test that messages persist in buffer when all processes exit."""
        messages = [b"persistent1", b"persistent2", b"persistent3"]

        # Writer writes and exits
        writer = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, messages)
        )
        writer.start()
        writer.join(timeout=5)
        self.assertFalse(writer.is_alive())

        # Small delay to ensure writer fully exits
        time.sleep(0.1)

        # New reader should still be able to read messages
        result_queue = multiprocessing.Queue()
        reader = multiprocessing.Process(
            target=reader_process,
            args=(self.buffer_name, len(messages), result_queue)
        )
        reader.start()
        reader.join(timeout=5)

        received = result_queue.get(timeout=1)
        self.assertEqual(received, messages, "Messages not persisted after writer exit")

    def test_concurrent_writes_sequential_reads(self):
        """Test multiple writers can write without corruption."""
        writer1_msgs = [b"w1_msg1", b"w1_msg2", b"w1_msg3"]
        writer2_msgs = [b"w2_msg1", b"w2_msg2", b"w2_msg3"]
        total_expected = len(writer1_msgs) + len(writer2_msgs)

        # Start both writers
        writer1 = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, writer1_msgs, 0.01)
        )
        writer2 = multiprocessing.Process(
            target=writer_process,
            args=(self.buffer_name, writer2_msgs, 0.01)
        )

        writer1.start()
        time.sleep(0.05)  # Stagger slightly
        writer2.start()

        writer1.join(timeout=5)
        writer2.join(timeout=5)

        # Read all messages
        result_queue = multiprocessing.Queue()
        reader = multiprocessing.Process(
            target=reader_process,
            args=(self.buffer_name, total_expected, result_queue, 10)
        )
        reader.start()
        reader.join(timeout=10)

        received = result_queue.get(timeout=1)

        # We should get all messages, though order may vary due to concurrency
        self.assertEqual(len(received), total_expected, "Not all messages from concurrent writers received")

        # Verify all messages from both writers are present
        for msg in writer1_msgs + writer2_msgs:
            self.assertIn(msg, received, f"Missing message: {msg}")


if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest

from packaging import version

from preseem import FakeNetworkMetadataModel, NetworkMetadataIpv6Settings, set_ipv6_settings, parse_ip_value

class TestIpv6(unittest.TestCase):
    """tests for IPv6 Settings class."""
    def setUp(self):
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        self.model = FakeNetworkMetadataModel(refs_by_type=True)
        asyncio.set_event_loop(self.loop)
        NetworkMetadataIpv6Settings._version_getter = self.fake_get_node_versions
        self.settings = None
        self.version_polls = 0
        self.node_versions = {}

    def tearDown(self):
        set_ipv6_settings(None)
        self._close()
        self.model.close()
        self.loop.close()

    async def fake_get_node_versions(self, telemetry_api):
        self.version_polls += 1
        return self.node_versions

    def _init(self, obj):
        self.settings = obj
        return self.loop.run_until_complete(obj.init())

    def _close(self):
        if self.settings:
            self.loop.run_until_complete(self.settings.close())

    def _sleep(self, sec):
        return self.loop.run_until_complete(asyncio.sleep(sec))

    def check_ipv6_enabled_state(self):
        """Return the value of the enabled reference state"""
        state_refs = self.model.refs.get('preseem_state') or {}
        ref = state_refs.get('ipv6_settings')
        return ref.attributes.get('ipv6_enabled') if ref else None

    def test_config_1(self):
        """Test configuration enabling ipv6"""
        settings = NetworkMetadataIpv6Settings(self.model, None, True)
        self._init(settings)
        self.assertEqual(settings.ipv6_enabled, True)
        self.assertEqual(self.version_polls, 0)

    def test_config_2(self):
        """Test configuration disabling ipv6"""
        settings = NetworkMetadataIpv6Settings(self.model, None, False)
        self._init(settings)
        self.assertEqual(settings.ipv6_enabled, False)

    def test_init_1(self):
        """The default initial state is ipv6 disabled."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self.assertEqual(self.check_ipv6_enabled_state(), None)
        self._init(settings)
        self.assertEqual(settings.ipv6_enabled, False)
        self.assertEqual(self.check_ipv6_enabled_state(), False)

    def test_init_2(self):
        """Once the reference is saved, it is used, no polling required."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self.node_versions = {'-node1': version.parse('1.11.1')}
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 1)
        self.assertEqual(settings.ipv6_enabled, True)
        self.assertEqual(self.check_ipv6_enabled_state(), True)
        self._close()
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 1)

    def test_init_3(self):
        """Polling is done if reference is ipv6 disabled."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self.node_versions = {'-node1': version.parse('0.11.1')}
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 1)
        self.assertEqual(settings.ipv6_enabled, False)
        self.assertEqual(self.check_ipv6_enabled_state(), False)
        self._close()
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 2)
        self.assertEqual(settings.ipv6_enabled, False)

    def test_poll_1(self):
        """Node versions of 1.11.1 or higher result in ipv6 enabled."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self.node_versions = {'-node1': version.parse('1.11.1')}
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 1)
        self.assertEqual(settings.ipv6_enabled, True)
        self.assertEqual(self.check_ipv6_enabled_state(), True)

    def test_poll_2(self):
        """Polling happens if ipv6 disabled, and stops when enabled."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        settings._POLL_INTERVAL = 1
        self.node_versions = {'-node1': version.parse('1.10.9')}
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(self.version_polls, 1)
        self._sleep(1)
        self.assertEqual(self.version_polls, 2)
        self.assertEqual(settings.ipv6_enabled, False)
        self.node_versions = {'-node1': version.parse('1.12.0')}
        self._sleep(1)
        self.assertEqual(self.version_polls, 3)
        self.assertEqual(settings.ipv6_enabled, True)
        self._sleep(1)
        self.assertEqual(self.version_polls, 3)
        self.assertEqual(self.check_ipv6_enabled_state(), True)

    def test_poll_3(self):
        """If there is a mix of versions, minimum version is used."""
        settings = NetworkMetadataIpv6Settings(self.model, None)
        self.node_versions = {'-node1': version.parse('1.11.1'), '-node2': version.parse('1.11.0')}
        self._init(settings)
        self._sleep(0.1)
        self.assertEqual(settings.ipv6_enabled, False)

    def test_parse_ipv4_1(self):
        """An IPv4 subnet is parsed as a subnet if ipv6 is enabled."""
        self.assertEqual(set(parse_ip_value('192.168.0.0/30')), set(['192.168.0.0', '192.168.0.1', '192.168.0.2']))
        set_ipv6_settings(NetworkMetadataIpv6Settings(self.model, None, ipv6_enabled=True))
        self.assertEqual(parse_ip_value('192.168.0.0/30'), ['192.168.0.0/30'])


    def test_parse_ipv4_2(self):
        """An IPv4 range is parsed as a subnet if ipv6 is enabled."""
        self.assertEqual(parse_ip_value('192.168.0.0-3'), ['192.168.0.0', '192.168.0.1', '192.168.0.2', '192.168.0.3'])
        set_ipv6_settings(NetworkMetadataIpv6Settings(self.model, None, ipv6_enabled=True))
        self.assertEqual(parse_ip_value('192.168.0.0-3'), ['192.168.0.0/30'])

    def test_parse_ipv6_1(self):
        """An IPv6 subnet is supported whether or not ipv6 is enabled."""
        self.assertEqual(parse_ip_value('2620:149:a43:301::/64'), ['2620:149:a43:301::/64'])
        set_ipv6_settings(NetworkMetadataIpv6Settings(self.model, None, ipv6_enabled=True))
        self.assertEqual(parse_ip_value('2620:149:a43:301::/64'), ['2620:149:a43:301::/64'])
        # quick normalization check
        self.assertEqual(parse_ip_value('2620:0149:a43:301::/64'), ['2620:149:a43:301::/64'])
